import pathlib
import dolfyn
from dolfyn.data.time import num2date
import warnings
import sys
import zipfile
from datetime import datetime
import numpy as np
import pandas as pd

thisdir = pathlib.Path(__file__).parent

if dolfyn.__version__ != '0.11.2':
    warnings.warn("This dataset was developed with DOLfYN Version 0.11.2. "
                  "There are no gaurantees it will work with other version.")

if sys.version_info.major != 3 or sys.version_info.minor < 7:
    warnings.warn(
        "This dataset was developed with Python 3.7. "
        "There are no gaurantees it will work with other version.")


def date2str(dt, format='ISO'):
    """Convert a datetime object, or array of them, to an array of strings.
    format can be 'ISO' (default, e.g.: 2020-03-20T14:28:23.382748),
    or a strftime format string.
    """
    if isinstance(dt, datetime):
        if format == 'ISO':
            return dt.isoformat()
        else:
            return dt.strftime(format)
    return np.array([date2str(d) for d in dt])
    

    
def to_csv(dat, fname, info_vars=[]):
    """Write the data to <fname>.csv.zip files

    dat must have the following variables:
      - z: which defines the depth bins
      - vel: the velocity vector
      - mpltime: the timestamps, which will be converted to strings

    The data object must be in 'earth' coords.

    The info_vars is a list of tuples [(outname, dat_field_name), ...].
    
    """

    # Make sure data is in earth
    assert dat.props['coord_sys'].lower() == 'earth'

    # Strip '.csv' if necessary.
    if fname.lower().endswith('.csv'):
        fname = fname[:-4]

    full_name = fname # Includes directory
    fname = fname.rsplit('/', 1)[-1] # Doesn't
        
    # Create the zipfile
    with zipfile.ZipFile(full_name + '.csv.zip',
                         mode='w',
                         compression=zipfile.ZIP_DEFLATED) as zfile:

        # Write out velocity components, each in its own .csv
        names = ['EAST', 'NORTH', 'UP', 'ERROR']
        time = date2str(num2date(dat.mpltime), format='ISO')
        for idx in range(4):
            df = pd.DataFrame(data=dat['vel'][idx].T,
                              index=time,
                              columns=dat['z'])
            zfile.writestr(
                fname + '.' + names[idx] + '.csv',
                df.to_csv(float_format='%0.4f'))

        # Write out "INFO.csv" file
        if len(info_vars) > 0:
            cols = []
            data = []
            for ky, datname in info_vars:
                cols.append(ky)
                data.append(dat[datname])
            df = pd.DataFrame(data=np.array(data).T,
                              index=time,
                              columns=cols,
                              copy=True)
            if 'Long' in df:
                df['Long'] = df['Long'].map(lambda x: '%.7f' % x)
            if 'Lat' in df:
                df['Lat'] = df['Lat'].map(lambda x: '%.7f' % x)
            zfile.writestr(fname + '.INFO.csv',
                           df.to_csv(float_format='%0.2f'))


def adv_to_csv(dat, fname, info_vars=[]):

    # Make sure data is in earth
    assert dat.props['coord_sys'].lower() == 'earth'
        
    time = date2str(num2date(dat.mpltime), format='ISO')
    df = pd.DataFrame(data={'East vel [m/s]': dat['vel'][0],
                            'North vel [m/s]': dat['vel'][1],
                            'Up vel [m/s]': dat['vel'][2], },
                      index=time)
    if len(info_vars) > 0:
        for nm, dname in info_vars:
            df[nm] = dat[dname]
    df.to_csv(fname, float_format='%0.4f')
