'''
GOAL: Process the Western Passage 2017 stable tidal turbulence mooring (STTM) data.
SMB stands for StableMoor Buoy, which is the main part of the STTM.
Use the starboard Nortek Vector (called ADVstar or ADV) and downward-looking Teledyne
RDI Workhorse ADCP (called SMBwkhs or RDI) data to fully motion correct
the ADVstar 16 Hz data, then calculate 5-min means for ADVstar and 
SMBwkhs. 

OVERVIEW OF STEPS 
1. Import binary files from ADVstar and SMBwkhs.
    Read the binary files, crop and clean them, and save .h5 files of ADV and RDI
    in ETU coordinates.
2. Timesync ADVstar and SMBworkhorse timestamps. Done externally then imported here.
3. Motion correct the ADVstar data. 
4. Calculate 'rotational' motion of ADP head and add it to datADvs.orient, call it ulow_rot
   Interpolate the 0.66 Hz RDI bt_vel and bt_range up to 16 Hz to match the datADVs data.
   Call them datADVs.orient['ulow_bt'] and datADVs.orient['ulow_bt_range']
   Calc ulow by: datADVs.orient['ulow'] = low pass filter (datADVs.orient['ulow_bt'] + datADVs.orient['ulow_rot'])
   Update rotate vars to include ulow and ulow_rot.
5. Calculate the SMB water depth using the depth data from the upward-looking Teledyne RDI V20 Sentinel ADCP . 
   The V20 was upward-looking on the SMB, but the data is only being used to calculate depth of the SMB.
6. Create vel = vel_orig + ulow  (vel = velraw + velrot + velacc - ulow_bt + ulow_rot)
7. Use TurbBinner to calculate 5-minute means for ADV in ETU (East, True North, Up)
8. Use TurbBinner to calculate 5-minute means for SMBworkhorse in ETU (East, True North, Up)
9. Import binary file from port Vector.
10. Export some of the the Starboard Vector variables to .csv
'''

#%%
import numpy as np
import datetime
import matplotlib
matplotlib.use('TKAgg') 
import matplotlib.dates as mdates
from matplotlib import pyplot as plt 
import scipy.io as sio
import scipy.signal as sig
import dolfyn
# See https://lkilcher.github.io/dolfyn/ for more information
import dolfyn.adv.api as avm
from dolfyn.tools.misc import fillgaps
from dolfyn.rotate.base import orient2euler
import dolfyn.data.velocity as dvel
import base

# Turn on interactive plotting
plt.ion()

#######
# USER NEEDS TO DEFINE FILEPATHS
fname_adv = 'WP2017_STTM_ADVstbd'
fname_rdi = 'WP2017_STTM_RDI'
fname_adv_port = 'WP2017_STTM_ADVport'

datdir = base.thisdir
figdir = base.thisdir / 'figs'
OUTDIR = base.thisdir

#######
# Seed the figure numbering
fignum = 1
dpi = 200

#%%
#######################################################################
# 1. Import ADV starboard and (StableMoor Buoy) SMB down-looking RDI workhorse
#    binary files. Read the binary files, clean them, and save .h5 files of
#    ADV and RDI in ETU coordinates. 
#######################################################################


def read_adv_stbd(outdir=OUTDIR):
    #######
    # Read in binary file from ADVstar and do preliminary DOLfYN processing. 
    # See dolfyn/io/nortek.py for more information.
    print('Reading the ADV .VEC file...')
    dat = avm.read(str(datdir / (fname_adv + '.VEC')), userdata=False) 
    print('Done reading the ADV .VEC file.')
    # ADV has no raw p,r,h because it has an IMU.
    # ADV does not have a userdata.json file at this time.

    #######
    # Define body2head information specific to this project.
    # **** Tag #130382 at front on star
    # **** Y = 31, x=-5.5, z=8.5 (inches)
    # The ADV body 'notch' faces nose of SMB, +x is forward
    # and since the head end was up, +z is down, so + y is to starboard.
    # The ADVstar body was mounted in the forwardmost ADV well.
    # Y = 31, x=-5.5, z=8.5 (inches) makes sense; to get from the ADVs body to the 
    # ADVs head, go back along the SMB centerline 5.5", then go to starboard 31", and down 8.5" to the head.
    # From dolfyn github web page, body2head_vec is the 3-element vector that specifies the position of the
    # ADV head in the ADV pressure case’s coordinate system
    dat.props['body2head_rotmat'] = np.eye(3)
    dat.props['body2head_vec'] = np.array([ -5.5, 31, 8.5]) * 0.0254

    #######
    # Set the declination.
    # https://www.ngdc.noaa.gov/geomag/calculators/magcalc.shtml#declination
    # wp SMB May 2017 = -16.74 deg if we define it according to 
    # https://lkilcher.github.io/dolfyn/data-structure.html
    # where 'The magnetic declination where the measurements were made
    # (in degrees that magnetic North is right of True North).'
    # At Western Passage, magnetic North is 16.74 deg to the LEFT of true North.
    # STTM location 44deg 55.259 N, 66deg 59.353 W
    # Since dat is currently in inst-frame, it will stay that way.
    # However, orientmat will be adjusted for declination so
    # later, when we rotate2 earth, resulting dat will be in ETU rather
    # than ENU.
    dat.set_declination(-16.74) 

    #######
    # Calculate and save DOLfYN-defined h,p,r
    odat = dat.orient
    (odat['heading'], odat['pitch'], odat['roll']) = orient2euler(odat['orientmat'])
    # Note dat.orient.raw will have heading, pitch, roll fields that are all 0s
    # because this ADV has an IMU rather than internal compass and tilt sensors.

    # This isn't used by dolfyn, but it may be useful and so we'll 
    # store it here if needed. It is the rotation matrix from the
    # body to the StableMoor buoy reference frame (i.e., it flips y and z)
    dat.props['body2SMB_rotmat'] = np.array([[1, 0, 0],
                        [0, -1, 0],
                        [0, 0, -1]])

    #######
    # Save
    print('Saving ...read_from_VEC_INST_d112.h5')
    dat.to_hdf5(str(outdir / (fname_adv + '_read_from_VEC_INST_d112.h5')))
    print('Done saving ...read_from_VEC_INST_d112.h5')
    # d112 = dolfyn v0.11.2

    #%%


def clean_adv_stbd(outdir=OUTDIR):
    #######################################################################
    # Clean ADVstar using the Goring+Nikora 2002 despiking method, with Wahl 2003
    # correction

    #######
    # Load data
    dat = dolfyn.load(str(outdir / (fname_adv + '_read_from_VEC_INST_d112.h5')))

    #######
    # Plot beginning of time series to figure out where to crop
    # the beginning of the data set (instrument in air or being put into 
    # position). Zoom in and identify the time when velocities start to look 
    # reasonable.

    # fignum = 1
    # fig = plt.figure(fignum, figsize=[8, 8])
    # fig.clf()
    # fig, axs = plt.subplots(1,1,num=fignum)

    #     for ivel in range(3):
    #         axs.plot(dat.mpltime[150000:200000], dat.vel[ivel][150000:200000], linestyle = 'none', marker = '.')

    start_advs_dn = 736473.6115
    start_advs_ds = mdates.num2date(start_advs_dn); print(start_advs_ds)
    # 2017-05-24 14:40:33.600000+00:00

    # Plot the end of the time series to figure out where to crop it.        
    # fignum = fignum + 1
    # fig = plt.figure(fignum, figsize=[8, 8])
    # fig.clf()
    # fig, axs = plt.subplots(1,1,num=fignum)

    #     for ivel in range(3):
    #         axs.plot(dat.mpltime[dat.mpltime.shape[0] - 400000:-1], dat.vel[ivel][dat.mpltime.shape[0] - 400000 :-1], linestyle = 'none', marker = '.')

    end_advs_dn = 736480.6070
    end_advs_ds = mdates.num2date(end_advs_dn); print(end_advs_ds)
    # 2017-05-31 14:34:04.800000+00:00 

    #######
    # Crop the raw data.
    # Before cropping dat.mpltime.shape (10022711,)
    trange = [start_advs_dn, end_advs_dn]
    tinds = (trange[0] <= dat.mpltime) & (dat.mpltime <= trange[-1])
    dat = dat.subset[tinds]  # dat.mpltime.shape (9593807,)

    #######
    # Plot cropped raw velocities, then keep plotting over them as the data is cleaned.
    fignum = 101
    fig = plt.figure(fignum, figsize=[8, 12])
    fig.clf()
    fig, axs = plt.subplots(3,1,num=fignum, sharex=True)

    for ivel in range(3):
        axs[ivel].plot(dat.mpltime, dat.vel[ivel], linestyle = 'none', marker = '.', color = 'k')

    axs[0].set_xlabel('time')        
    axs[0].set_ylabel('u')
    axs[1].set_ylabel('v')
    axs[2].set_ylabel('w')

    #######
    # Clean with GN2002
    # The avm.clean.GN2002(dat.u) part adjusts the u vel within dat.
    # The bads_u part grabs the output of GN2002 and reports the number of 
    # values that have been interpolated over.
    bads_u = avm.clean.GN2002(dat.u) # bads_u.sum() = 3346
    prcnt_u_cleaned = bads_u.sum()/dat.u.shape[0] * 100 ; print(prcnt_u_cleaned) # 0.03 %
    bads_v = avm.clean.GN2002(dat.v) # bads_v.sum() = 2421
    prcnt_v_cleaned = bads_v.sum()/dat.v.shape[0] * 100 ; print(prcnt_v_cleaned) # 0.03 %
    bads_w =avm.clean.GN2002(dat.w) # bads_w.sum() = 4146
    prcnt_w_cleaned = bads_w.sum()/dat.w.shape[0] * 100 ; print(prcnt_w_cleaned) # 0.04 %

    #######
    # Add cleaned data to the previous plot
    for ivel in range(3):
        axs[ivel].plot(dat.mpltime, dat.vel[ivel], linestyle = 'none', marker = '.', color = 'y')

    axs[0].set_title('ADVstar cropped raw vel (black) and cleaned with GN2002 (yellow)')
    axs[0].set_ylabel('x')
    axs[1].set_ylabel('y')
    axs[2].set_ylabel('z')
    axs[-1].xaxis.set_major_locator(mdates.DayLocator(interval=1)) 
    axs[-1].xaxis.set_minor_locator(mdates.HourLocator(interval=12))  
    axs[-1].xaxis.set_major_formatter(mdates.DateFormatter('%-m/%-d')) 
    fig.savefig(str(figdir / 'ADVs_cleaned_GN2002.png') )

    #######
    # To match the SMBworkhorse binary file coordinate system, 
    # rotate to ETU.
    dolfyn.rotate2(dat, 'earth', inplace=True)
    # Save
    print('Saving clnETU_d112.h5...')
    dat.to_hdf5(str(outdir / (fname_adv + '_clnETU_d112.h5')))
    print('Done saving ...clnETU_d112.h5')

    #%%

def read_RDI(outdir=OUTDIR):
    #######################################################################
    # SMBworkhorse 

    #######
    # Read in the binary file.
    print('Reading the .000 SMBwkhs file')
    dat = dolfyn.read(str(datdir / (fname_rdi + '.000')), userdata=True)
    print('Done reading the .000 SMBwkhs file')
    # read.rdi includes _set_rdi_declination 
    # declination = -16.74 in userdata.json

    #######
    # Calc and save DOLfYN-defined h,p,r
    odat = dat.orient
    (odat['heading'], odat['pitch'], odat['roll']) = orient2euler(odat['orientmat'])

    #######
    # Save
    print('Saving SMBwkhs_read_from_000_ETU_d112.h5...')
    dat.to_hdf5(str(outdir / (fname_rdi + '_read_from_000_ETU_d112.h5')))
    print('Done saving ...read_from_000_ETU_d112.h5')

    #%%
    # #####################################################################
    # Clean the SMBwkhs data.


def clean_RDI(outdir=OUTDIR):
    #######
    # Load data.
    dat = dolfyn.load(str(outdir / (fname_rdi + '_read_from_000_ETU_d112.h5')))

    #######
    # Plot the velocity to decide where to crop the data at the beginning
    # of the time series.
    # fignum = fignum + 1
    # fig = plt.figure(fignum, figsize=[8, 8])
    # fig.clf()
    # fig, axs = plt.subplots(1,1,num=fignum)

    # for ivel in range(3):
    #     axs.plot(dat.mpltime[0:15000], dat.vel[ivel][1][0:15000], linestyle = 'none', marker = '.')

    start_rdi_dn = 736473.610
    # start_rdi_ds = mdates.num2date(start_rdi_dn); print(start_rdi_ds)
    # 2017-05-24 14:38:24+00:00

    # Look at the end of the time series to figure out where to crop it.        
        # fignum = fignum + 1
        # fig = plt.figure(fignum, figsize=[8, 8])
        # fig.clf()
        # fig, axs = plt.subplots(1,1,num=fignum)

        #     for ivel in range(3):
        #         axs.plot(dat.mpltime[dat.mpltime.shape[0] - 15000:-1], dat.vel[ivel][1][dat.mpltime.shape[0] - 15000:-1], linestyle = 'none', marker = '.')

    # Doesn't need to be cropped at end.
    # However, after going through later steps in this script, 
    # there are NaN at the end of the time series that I want 
    # to crop out, so change 
    # end_rdi_dn = dat.mpltime[-1] to dat.mpltime[-9918]
    end_rdi_dn = dat.mpltime[-9918]
    # end_rdi_ds = mdates.num2date(end_rdi_dn); print(end_rdi_ds)
    # 2017-05-31 14:35:58.006780+00:00

    #######
    # Crop the raw data.
    # Before cropping dat.mpltime.shape (419748,)
    trange = [start_rdi_dn, end_rdi_dn]
    tinds = (trange[0] <= dat.mpltime) & (dat.mpltime <= trange[-1])
    dat = dat.subset[tinds]  # dat.mpltime.shape (403103,)

    #######
    # Bottom Tracking (BT) percent goods
    # The percent good fields are populated differently than you
    # might expect (i.e. that they mean the % of good values) because the 
    # coordinate system was set to Earth.
    # Field 1 is the % of good 3-beam solutions. Since the best solutions
        # are 4 beam solutions (Field 4), high values here
        # are good if Field 4 values are low.
    # Field 2 is the % of transformations rejected because the error
        # velocity was over the threshold (2000 mm/s)
    # Field 3 is the % where more than 1 beam are bad in a bin.
    # Field 4 is the % of 4-beam solutions. Best data.

    #######
    # BT amplitude and correlation. 
    # WorkHorse Commands and Output Data Format.pdf 
    # default minimum threshold for amplitude is 30. Use to filter bt data.
    # default minimum threshold for correlation is 220. Use to filter bt data.
    # For this dataset, it looks to be a fair threshold.

    # fignum = fignum + 1
    # fig = plt.figure(fignum, figsize=[8, 10])
    # fig.clf()
    # fig, axs = plt.subplots(6,1,num=fignum, sharex=True)

    # for ivel in range(2):
    #     axs[0].plot(dat.mpltime, dat.signal.bt_ampl[ivel], linestyle = 'none', marker = '.')
    #     axs[0].set_xlabel('time')
    #     axs[0].set_ylabel('bt_ampl')

    # for ivel in range(3):
    #     axs[1].plot(dat.mpltime, dat.signal.bt_corr[ivel], linestyle = 'none', marker = '.')
    #     axs[1].set_xlabel('time')
    #     axs[1].set_ylabel('bt_corr')

    # axs[2].plot(dat.mpltime, dat.bt_vel[0], linestyle = 'none', marker = '.', color = 'k')
    # axs[2].set_xlabel('time')
    # axs[2].set_ylabel('bt u')

    # axs[3].plot(dat.mpltime, dat.bt_vel[1], linestyle = 'none', marker = '.', color = 'k')
    # axs[3].set_xlabel('time')
    # axs[3].set_ylabel('bt v')

    # axs[4].plot(dat.mpltime, dat.bt_vel[2], linestyle = 'none', marker = '.', color = 'k')
    # axs[4].set_xlabel('time')
    # axs[4].set_ylabel('bt w')

    # axs[5].plot(dat.mpltime, np.nanmean(dat.bt_range, axis=0), linestyle = 'none', marker = '.', color = 'r')
    # axs[5]. set_ylabel('bt range mean (m)')

    # axs[0].xaxis.set_major_locator(mdates.DayLocator(interval=1))  # Major tick every 1 month
    # axs[0].xaxis.set_minor_locator(mdates.HourLocator(interval=12))  # Minor tick every 1 day
    # axs[0].xaxis.set_major_formatter(mdates.DateFormatter('%-m/%-d'))  # Label the major ticks
    # axs[0].set_title('SMBworkhorse after cropping, before filtering')
    # fig.savefig(str(figdir / 'SMBwkhs_bt_amp_corr.png'))
    # DON'T CLOSE THIS PLOT YET.

    # Copy the variable in order to compare it with original later.
    dat['bt_vel_gd'] = dat.bt_vel.copy() 

    # How many NaN are in each vel component before any filtering?
    # a = np.isnan(dat.bt_vel_gd[0]).sum(); print(a) # 0
    # b = np.isnan(dat.bt_vel_gd[1]).sum(); print(b) # 0
    # c = np.isnan(dat.bt_vel_gd[2]).sum(); print(c) # 0
    # d = np.isnan(dat.bt_vel_gd[3]).sum(); print(d) # 0

    #######
    # FIRST, NaN bt_vel values corresponding to bt_corr < 220 & bt_ampl < 30
    for ivel in range(3):
        ibad = ((dat.signal.bt_corr[ivel] < 220) | (dat.signal.bt_ampl[ivel] < 30))
        dat.bt_vel_gd[ivel, ibad] = np.nan

    # How many NaN are in each bt_vel component? Full length of time series is 403103 events
    # a = np.isnan(dat.bt_vel_gd[0]).sum(); print(a) # 26
    # b = np.isnan(dat.bt_vel_gd[1]).sum(); print(b) # 66
    # c = np.isnan(dat.bt_vel_gd[2]).sum(); print(c) # 31
    # d = np.isnan(dat.bt_vel_gd[3]).sum(); print(d) # 0

    #######
    # SECOND, NaN more using percent good
    ibad = (((dat.signal.bt_perc_gd[0] + dat.signal.bt_perc_gd[3]) < 90) | (dat.signal.bt_perc_gd[2] > 20))

    for ivel in range(3):
        dat.bt_vel_gd[ivel, ibad] = np.nan

    # How many NaN are in each bt_vel component? A few more than before.
    # a = np.isnan(dat.bt_vel_gd[0]).sum(); print(a) # 37
    # b = np.isnan(dat.bt_vel_gd[1]).sum(); print(b) # 72
    # c = np.isnan(dat.bt_vel_gd[2]).sum(); print(c) # 39
    # d = np.isnan(dat.bt_vel_gd[3]).sum(); print(d) # 0

    # Add to the previous figure to make sure those values are gone. Good.
    # axs[2].plot(dat.mpltime, dat.bt_vel_gd[0], linestyle = 'none', marker = '.', color = 'b')
    # axs[2].set_xlabel('time')
    # axs[2].set_ylabel('bt u')

    # axs[3].plot(dat.mpltime, dat.bt_vel_gd[1], linestyle = 'none', marker = '.', color = 'g')
    # axs[3].set_xlabel('time')
    # axs[3].set_ylabel('bt v')

    # axs[4].plot(dat.mpltime, dat.bt_vel_gd[2], linestyle = 'none', marker = '.', color = 'y')
    # axs[4].set_xlabel('time')
    # axs[4].set_ylabel('bt w')

    # axs[0].set_title('SMBworkhorse after cropping, black = before filtering, colors in vel = after filters')
    # fig.savefig(str(figdir / 'SMBwkhs_bt_amp_corr_after_filt.png'))

    #######
    # Interpolate the Nan'd bt_vel_gd and plot to check.
    # dat.bt_vel_gd.shape  #  (4, 403103)
    fillgaps(dat.bt_vel_gd, dim=1, extrapFlg=False)  
    # If extrapFlg=False, there could be NaN after fillgaps if
    # there are NaN at the end of the time series.
    # That's why I cropped out those NaN at the end of the
    # time series earlier.

    #######
    # Assign dat.bt_vel_gd to dat.bt_vel, where it belongs
    dat['bt_vel'] = dat['bt_vel_gd']
    # Get rid of unneeded variables  
    dat.pop('bt_vel_gd')  

    #######
    # At this point, the SMBworkhorse bt_vel data has been 'cleaned', but the water column velocity data
    # has not. 

    # Broadband Primer pdf indicates a corr of 128 or greater indicates
    # a good signal-to-noise ratio. We'll throw out a ton of data if the threshold is increased to 128 (62.7%)
    # instead of 15.7% (compared to baseline 14.3% NaNs) when corr < 64 is used.
    # but perhaps that's okay since so many of the bins are close to the bottom so often. 
    # Look at velocities and corr and see what a reasonable threshold for this data is. 

    #######
    # See how much data gets cut when this threshold is employed: dat.signal.corr[ivel] < 64 
    # Assign temporary variable.
    dat['vel_gd'] = dat.vel.copy()

    for ivel in range(3):
        ibad = (dat.signal.corr[ivel] < 64) 
        dat.vel_gd[ivel, ibad] = np.nan

    iNotnan2 = ~np.isnan(dat.vel_gd); print(iNotnan2.sum()) # 
    inan2 = np.isnan(dat.vel_gd); print(inan2.sum()) # 
    tots2 = iNotnan2.sum() + inan2.sum()
    prcnt_nan2 = inan2.sum() / tots2 * 100  # When the threshold is < 128, 62.7% of the data is excluded.
                                            # When < 64 is used, 15.7% excluded  
                                            # compared to the baseline 14.3% 
    print(prcnt_nan2)


    #######
    # So now the data is filtered for low corr, but still exists for unrealistic ranges. 
    #  The Broadband Primer_Dec06.pdf, pg. 38. Rmax = D cos(theta) where
        # Rmax is maximum range for acceptable data, D is distance from ADCP to surface (or bottom),
        # theta is angle of beam relative to vertical (in this case 20 degrees). 
        # cos(theta) = .94

    for ivel in range(3):
        goods = dat.range[:, None] < dat.bt_range[ivel, :] * 0.9 # Conservative. 90% instead of 94%
        dat.vel_gd[ivel, ~goods] = np.NaN

    #######
    # Assign dat.vel_gd to dat.vel and pop dat.vel_gd
    dat.vel = dat.vel_gd
    dat.pop('vel_gd')

    #######
    # Finally, IEC TS 62600-201 requires data from bins
    # that do not have good values at least 95% of the time to be rejected. 
    # For an up-looking ADCP, this effectively removes the bins 
    # associated with high tide water levels, so all time chunks
    # can be compared apples-to-apples regardless of tidal stage.
    # Here, since the SMB is down-looking, the high tide bins are not an issue.
    # Here, the SMB flies at different heights, so data from dat.range[0] is 
    # always 2.09 m from the transducer, but it is not always the same distance
    # from the bottom.
    # What is an appropriate way to only use bins that have a lot of good data?
    # I suggest the following approach:
    # Make a histogram of the the height above the bottom and NaN vel data
    # collected when the SMB was not at a common hab (e.g. 9.5 - 10.5 m as in 
    # the figure below).

    # Create height above bottom (hab) variable
    dat['hab'] = np.nanmean(dat.bt_range, axis = 0)

    # Histogram of bt_range and hab
    # fignum = fignum + 1
    # fig = plt.figure(fignum, figsize=[9, 5])
    # fig.clf()
    # fig, axs = plt.subplots(1, 1, num=fignum)
    # axs.hist(dat.bt_range[0],bins = 100, label = '1')
    # axs.hist(dat.bt_range[1],bins = 100, label = '2')
    # axs.hist(dat.bt_range[2],bins = 100, label = '3')
    # axs.hist(dat.bt_range[3],bins = 100,label = '4')
    # axs.hist(dat.hab, bins = 100, label = 'avg')
    # axs.legend()
    # axs.set_xlabel('height above bottom [m]')
    # axs.set_title('SMB workhorse hab histogram', loc='center')
    # fig.savefig(str(figdir / 'SMBwkhs_hab_hist.png'), dpi=dpi)

    dat['vel_gd'] = dat['vel'].copy()

    # Since hist of hab shows most between 9.5 and 10.5, use this instead of the
    # 95%-of-the-time-good criteria.
    # So good data is restricted to when SMB was flying between 9.5 and 10.5 m above the bottom
    igd =  (dat['hab'] > 9.5) & (dat['hab'] < 10.5) # Boolean. 
    # What percentage of all hab are in igd?
    prcnt_gdhab = igd.sum() / dat['hab'].shape[0] * 100; print(prcnt_gdhab) # 82.8%
    dat.vel_gd[:, :, ~igd] = np.NaN

    #######
    # Assign dat.vel_gd to dat.vel, where it belongs
    dat['vel'] = dat['vel_gd']
    # Get rid of unneeded variables  
    dat.pop('vel_gd') 

    # Load the .csv with RDI data synced to the ADVstarboard time, call it datRDI.advtime
    dat['advtime'] = np.loadtxt(str(datdir / 'WP2017_STTM_ADVstbd_corrected_RDI_time_1H_Bins.csv'), delimiter=',', usecols= 1)

    #######
    # Update the dat.props to use the correct fs, which is 0.66 Hz
    dat.props['fs'] = np.round(1/ 1.5, decimals=3) 

    #######
    # Save
    print('Saving SMBwkhs_clnETU_d112.h5...')
    dat.to_hdf5(str(outdir / (fname_rdi + '_timesynced2ADVs_clnETU_d112.h5')))
    print('Done saving..._clnETU_d112.h5')


def correct_motion(outdir=OUTDIR):
    # To calculate vellow:
    # 3. Motion correct the ADVstar data 
    # 4. Calculate 'rotational' motion of ADP head and add it to datADvs.orient, call it ulow_rot
    # 5. Interpolate the 0.66 Hz RDI bt_vel and bt_range up to 16 Hz to match the datADVs data.
    #    Call them datADVs.orient['ulow_bt'] and datADVs.orient['ulow_bt_range']
    # 6. Calc ulow by: datADVs.orient['ulow'] = low pass filter (- datADVs.orient['ulow_bt'] + datADVs.orient['ulow_rot'])
    # 7. Update rotate vars to inclue ulow and ulow_rot, 
    #    so they will be rotated when data is rotated to principal axis (not done here).
    # 8. Create new vel = vel_orig + ulow  (vel = velraw + velrot + velacc - ulow_bt + ulow_rot)

    #######
    # Load the data. Both datADVs and datRDI are in ETU.
    datADVs = dolfyn.load(str(outdir / (fname_adv + '_clnETU_d112.h5')))
    datRDI = dolfyn.load(str(outdir / (fname_rdi + '_timesynced2ADVs_clnETU_d112.h5')))
    filt_freq = 0.2 # Set to 0.2 after exploring results.

    # FYI, from motion.py re: avm.motion.CorrectMotion
    #     accel_filtfreq : float
    #   the frequency at which to high-pass filter the acceleration
    #   signal to remove low-frequency drift. (default: 0.03 Hz)

    #     vel_filtfreq : float (optional)
    #   a second frequency to high-pass filter the integrated
    #   acceleration.  (default: 1/3 of accel_filtfreq)

    #######
    # Motion correct the ADVs data
    mc = avm.motion.CorrectMotion(accel_filtfreq = filt_freq)
    # Although /adv/motion.py says to use 'correct_motion' function instead
    # use this class until that function's accuracy is verified.
    # Since we have independent measure of vellow from SMBworkhorse, don't use the
    # default accel_filtfreq of 0.03 but rather use 0.1 (or 0.2 as is done here)
    print("Motion correcting...")
    mc(datADVs) 
    print('Done motion correcting')

    #######
    # Calculate 'rotational' motion of ADP head and add it to datADvs.orient
    motcalc = avm.motion.CalcMotion(datADVs)
    datADVs.orient['ulow_rot'] = motcalc.calc_velrot(np.array([-0.63, 0, -0.86]), to_earth=False)
    # The vector provided as input to calc_velrot is meant to be the body2head vector FROM the ADVbody coordinate system
    # (origin where the body endcap meets the head cable) TO the ADV head. Within
    # calc_velrot, this vector is modified to represent the vector from the IMU to the ADV head (l^* in 
    # Kilcher et al. 2016 NREL manual and Kilcher et al. 2017 Turbulence...Part II):
            # From calc_velrot:
            # Correct for the body->imu distance.
            # The nortek_body2imu vector is subtracted because of
            # vector addition:
            # body2head = body2imu + imu2head
            # Thus:
            # imu2head = body2head - body2imu
            # vec = vec - get_body2imu(self.advo._make_model)[:, None]

    # In this instance, the body2head vector provided is the body2RDIface vector.
    # [-0.63, 0, -0.86] is distance (m) from IMU to RDI face.
    # ADVs body to RDI is -0.63 m (back along main axis of SMB)
    # The vertical distance from ADV body coordinate system to the face of the RDI is -0.86 m.
    # to_earth = False because datADVs after mc is already in earth coordinates.
    # Note that this used datADVs, so it's at 16 Hz.

    #######
    # Interpolate the 0.66 Hz RDI bt_vel up to 16 Hz to match the datADVs data. 
    bt_vel_16Hz = np.empty_like(datADVs.vel)
        # ulow_range = np.empty_like(datADVs.vel)
    for idx in range(3):
        bt_vel_16Hz[idx] = np.interp(datADVs.mpltime, datRDI.advtime, datRDI.bt_vel[idx])
                        # MCF: new_y = np.interp(new_x, old_x, old_y)
                        # So ulow is bt_vel being interpolated from the synced RDI time
                        # to the 16 Hz ADVs mpltime.
                        # And the fourth row of bt_vel is being dropped.

    #######
    # Low pass filter ulow with the same filt_freq
    # used in the mc of the ADVs data.
    # Apply the same filter as the ADV high pass filter used on the ADV.
    # Don't do the subtraction because we want the results of the low pass filter.

    # This is in adv/motion.py  line 108
    #  def _set_acclow(self, ):
    #         """
    #         """
    #         self.acclow = acc = self.accel.copy()
    #         if self.accel_filtfreq == 0:
    #             acc[:] = acc.mean(-1)[..., None]
    #         else:
    #             flt = sig.butter(1, self.accel_filtfreq / (self.advo['props']['fs'] / 2))
    #             for idx in range(3):
    #                 acc[idx] = sig.filtfilt(flt[0], flt[1], acc[idx])

    flt = sig.butter(2, filt_freq / datADVs['props']['fs'] / 2)

    # Before low-pass filtering bt_vel_16Hz, remove the rotational velocity of the RDI Workhorse about the
    # IMU. In the same way ADV vel = velraw + velrot, add the ulow_rot to bt_vel_16Hz to remove the motion.
    bt_vel_16Hz += datADVs.orient['ulow_rot']  

    # Next apply a 4th order, zero-phase (bidirectional) Butterworth filter to ulow to produce the 
    # 'ulow' to be used in the full motion correction of the ADV vel (later we SUBTRACT ulow from vel).
    datADVs.orient['ulow'] = sig.filtfilt(flt[0], flt[1], bt_vel_16Hz ) 

    #######
    # Update dat.props['rotate_vars'] to include the new variables.
    # This is needed when the coordinate system is rotated to principal axes (not done here).
    datADVs.props['rotate_vars'].update({'orient.ulow', 'orient.ulow_rot'})   

    #######
    # Interpolate the bt_range and hab_rdi, too, and put it on datADVs.
    tmp = np.empty_like(datADVs.vel)
    new_row = tmp.sum(0)[None,...]; new_row.shape
    datADVs['ulow_bt_range'] = np.append(tmp, new_row, 0); datADVs['ulow_bt_range'].shape

    for idx in range(4):
        datADVs['ulow_bt_range'][idx] = np.interp(datADVs.mpltime, datRDI.advtime, datRDI.bt_range[idx])
    # In the SMB workshorse plots, hab is calc'd from all 4 beam fields. 
    datADVs.props['rotate_vars'].update({'ulow_bt_range'})  

    datADVs['ulow_hab'] = np.interp(datADVs.mpltime, datRDI.advtime, datRDI.hab)

    #######
    #######
    # Calculate the water column depth at the STTM site, then calculate MLLW.

    #######
    # DOLfYN is unable to read the v20 file, so it was read in using MATLAB
    # and the 'rdradcp' function from RDADCP_Jan18v0 at https://www.eoas.ubc.ca/~rich/#RDADCP
    # and the data is in beam coordinates.
    # Load the v20 .mat file
    x = sio.loadmat(str(datdir / 'WP2017_STTM_v20.mat'))

    v20struc = x['A'] # v20struc.shape is (1, 1), just like the MATLAB structure                                                                            
    val = v20struc[0,0] 
    # These are inside A
    time = np.transpose(val['mtime'][0,:]) # shape (252606,)
    depth = np.transpose(val['depth'][0,:]) # shape (252606,)

    # Add these to the ADV dat object
    datADVs['time_v20'] = time - 366 # To adjust for MATLAB to Python time
    datADVs['depth_v20'] = depth 

    #######
    # Crop the v20 to match the ADV time, upsample the v20 to 16 Hz,
    # calc water depth as depth_v20 + hab_rdi + m_v20toWkhs

    igd = (datADVs.time_v20 >= np.min(datADVs.mpltime)) & (datADVs.time_v20 <= np.max(datADVs.mpltime))
    datADVs['time_v20'] = datADVs['time_v20'][igd]
    datADVs['depth_v20'] = datADVs['depth_v20'][igd]
    datADVs['depth_v20_16Hz'] = np.interp(datADVs['mpltime'],datADVs['time_v20'],datADVs['depth_v20'])

    #######
    # Estimate distance from pressure sensor on v20 to transducer face on Workhorse from SMB diagram and photos 
    v20toWkhs_m = 14 * 2.54 / 100
    datADVs['water_depth_est'] = datADVs['depth_v20_16Hz'] + datADVs['ulow_hab'] + v20toWkhs_m

    #######
    # Calculate MLLW
    # Grab each day individually.
    d1  = datADVs.water_depth_est[:1370543]  
    d2  = datADVs.water_depth_est[1370543:2741086] 
    d3  = datADVs.water_depth_est[2741086:4111629] 
    d4  = datADVs.water_depth_est[4111629:5482172] 
    d5  = datADVs.water_depth_est[5482172:6852715]
    d6  = datADVs.water_depth_est[6852715:8223258]
    d7  = datADVs.water_depth_est[8223258:9593801]

    dd = [d1, d2, d3, d4, d5, d6, d7]
    dd_min = np.nanmin(dd, axis = 1)
    datADVs['mllw'] = np.round(np.mean(dd_min), decimals = 1) # mllw = 53 m

    #######
    # Get rid of uneccessary values
    datADVs.pop('depth_v20')
    datADVs.pop('time_v20')

    #######
    # Save
    print('Saving ...d112_w_ulow_waterdepth_ETU.h5')
    datADVs.to_hdf5(str(outdir / (fname_adv + '_d112_w_ulow_waterdepth_ETU.h5')))
    print('Done saving ...d112_w_ulow_waterdepth_ETU.h5')


def avg_adv_stbd(outdir=OUTDIR):
    # Use TurbBinner to calculate 5-minute means of motion-corrected
    # ADV data in ETU. 

    #######
    # Load data.
    dat = dolfyn.load(str(outdir / (fname_adv + '_d112_w_ulow_waterdepth_ETU.h5')))

    #######
    # So many calcs act on dat.vel. Call the good one with ulow dat.vel and the orig one vel_orig
    # Currently dat['vel'] is only vel = velraw + velrot + velacc.
    # We want it to be vel = velraw + velrot + velacc - ulow
    dat['vel_orig'] = dat['vel']
    dat['vel'] = dat['vel_orig'] - dat.orient['ulow']

    #######
    # Define Doppler noise for all three velocity vectors.
    # Set to 0s if the quantities haven't been identified yet,
    # then rerun TurbBinner with values again to get the 'final answer' 5 min means.
    doppler_noise_ms = [0, 0, 0]
    dat['doppler_noise'] = doppler_noise_ms

    #######
    # Calculate 5-minute means in ETU
    bnr = avm.TurbBinner(5 * 60 * dat.props['fs'] , dat.props['fs'] )
    print("Averaging...")

    # Next step will produce sigma_Uh, stress, tke_vec, and 
    # bdat.Spec['omega'] and bdat.Spec['vel'] 
    bdat = bnr(dat) 
    print('Done with bnr(dat)')

    bdat['Spec.velraw'] = bnr.calc_vel_psd(dat['velraw'], noise=dat['doppler_noise'])
    bdat['Spec.velrot'] = bnr.calc_vel_psd(dat.orient['velrot'], noise=dat['doppler_noise'])
    bdat['Spec.velacc'] = bnr.calc_vel_psd(dat.orient['velacc'], noise=dat['doppler_noise'])
    bdat['Spec.ulow'] = bnr.calc_vel_psd(dat.orient['ulow'], noise=dat['doppler_noise']) 
    bdat['Spec.vel_orig'] = bnr.calc_vel_psd(dat['vel_orig'], noise=dat['doppler_noise'])
    bdat['Spec.vel'] = bnr.calc_vel_psd(dat['vel'], noise=dat['doppler_noise'])

    # For plotting horizontal motion
    noise_horiz =  (dat['doppler_noise'][0] + dat['doppler_noise'][0]) / 2
    bdat['Spec.Umag'] = bnr.psd(dat.U_mag, noise=noise_horiz) # This is U_mag for vel (with ulow)

    dat['U_mag_orig'] = np.sqrt(dat.vel_orig[0] ** 2 + dat.vel_orig[1] ** 2)
    bdat['Spec.Umag_orig'] = bnr.psd(dat['U_mag_orig'], noise=noise_horiz) # This is U_mag for vel_orig

    bdat['TI'] = bdat['sigma_Uh'] / bdat.U_mag
    
    #######
    # Save
    print('Saving ADV ... d112_ETU_5min.h5')
    bdat.to_hdf5(str(outdir / (fname_adv + '_d112_ETU_5min.h5')))
    print('Done saving ADV ...d112_ETU_5min.h5')


def avg_RDI(outdir=OUTDIR):
    # Calculate 5-minute means for the SMBwkhs data.

    #######
    # Load 0.66 Hz data that have been time-synced to ADV starboard clock and are in ETU.
    dat = dolfyn.load(str(outdir / (fname_rdi + '_timesynced2ADVs_clnETU_d112.h5')))

    #######
    # Earlier, I NaN'd vels like this
    # igd =  (dat['hab'] > 9.5) & (dat['hab'] < 10.5) # Boolean. 
    # dat.vel_gd[:, :, ~igd] = np.NaN
    # but now need to NaN these other variables before calculating averages.
    igd =  (dat['hab'] > 9.5) & (dat['hab'] < 10.5) 
    dat['bt_vel'][:, ~igd] = np.NaN
    dat['bt_range'][:, ~igd] = np.NaN
    dat['depth_m'][~igd] = np.NaN
    dat['hab'][~igd] = np.NaN
    dat.U_mag[:,~igd] = np.NaN

    #######
    # Adjust Umag for bt_vel
    dat.U_mag2 = dat.U_mag - np.sqrt(dat['bt_vel'][0] ** 2 + dat['bt_vel'][1] ** 2)

    #######
    # Since SMB was flying between 9.5 and 10.5 meters above bottom, the range bins
    # still don't represent the same distance from the bottom, but I will call the 
    # height above the bottom = 10 m for all of this data. 

    # If the bottom is 53.0 m deep relative to MLLW, 
    # and range is 2.09, 3.09, 4.09, 5.09, 6.09, 7.09, 8.09, 9.09, and 10.09 and those values
    # indicate the distance of the bin from the ADCP face, those range values are equivalent
    # to a DEPTH of 45.09, 44.09,...52.09, 53.09.
    dat['depth_wrt_mllw'] = dat['range'] + 43

    # Subset the data to only include vels associated with dat['depth_wrt_mllw'] up to 53.09.
    # Also crop the range and depth_wrt_mllw.
    # Upon further review later, it looks like there's a lot of bad data in the last two depth bins, so
    # crop at [:7] instead of [:9]]
    dat['vel'] = dat['vel'][:, :7, :]
    dat['range'] = dat['range'][:7]
    dat['depth_wrt_mllw'] = dat['depth_wrt_mllw'][:7] 
    dat.U_mag2 =  dat.U_mag2[:7,:]

    #######
    # Calculate 5-min means
    nbin = 5 * 60 / 1.5 # 5 min * 60 sec/min * 1 sample/1.5 sec = 200 samples per bin
    bnr = dvel.VelBinner(n_bin = nbin, fs = dat.props['fs'])
    # Since there are NaNs in the time series, use reshape instead of do_avg
    arr = bnr.reshape(dat.vel) # arr.shape  (4, 7, 2015, 200)4 vels, 7 bins, 2015 5 min chunks, each with 200 samples
    dat['vel_mean'] = np.nanmean(arr, axis = 3) #  dat['vel_mean'].shape (4, 7, 2015)
    arr2 = bnr.reshape(dat.mpltime)
    dat['mpltime_mean'] = np.nanmean(arr2, axis = 1) # shape (2015,)
    # What else needs to be averaged? bt_vel since vel = vel - bt_vel
    # arr3 = bnr.reshape(dat.bt_vel) # arr3.shape (4, 2015, 200)
    # dat['bt_vel_mean'] = np.nanmean(arr3, axis = 2) # dat['bt_vel_mean'].shape  (4, 2015)
    arr4 = bnr.reshape(dat.bt_range) # arr4.shape (4, 2015, 200)
    dat['bt_range_mean'] = np.nanmean(arr4, axis = 2) # dat['bt_range_mean'].shape  (4, 2015)
    arr5 = bnr.reshape(dat.depth_m) # arr5.shape (2015,200)
    dat['depth_m_mean'] = np.nanmean(arr5, axis = 1) # dat['depth_m_mean'].shape  (2015,)
    arr6 = bnr.reshape(dat.hab) # arr6.shape (2014,200)
    dat['hab_mean'] = np.nanmean(arr6, axis = 1) # dat['hab_mean'].shape  (2015,)
    arr7 = bnr.reshape(dat.U_mag2) # arr7.shape (7, 2015,200)
    dat['U_mag_mean'] = np.nanmean(arr7, axis = 2) # dat['U_mag_mean'].shape  (7, 2014,)

    # Save 
    print('Saving SMBwkhs ...d112_5min_ETU.h5')
    dat.to_hdf5(str(outdir / (fname_rdi + '_d112_5min_ETU.h5')))
    print('Done saving SMBwkhs ...d112_5min_ETU.h5')


def read_adv_port(outdir=OUTDIR):
    # Import the port Nortek Vector binary data into DOLfYN.
    # No additional processing is done.

    #######
    # Read in binary file from ADVport
    print('Reading the ADV port .VEC file...')
    dat = avm.read(str(datdir / (fname_adv_port + '.VEC')), userdata=False)
    print('Done reading the ADV port .VEC file')
    # ADV has no raw p,r,h because it has an IMU.
    # ADV does not have a userdata.json file at this time.

    #######
    # Define body2head information specific to this project.
    # **** Tag #130381 at port
    # **** Y=-31, X=5, Z=8.25 (inches) 
    # The ADV body 'notch' faces nose of SMB, +x is forward
    # and since the head end was up, +z is down, so + y is to starboard.
    # The ADVport body was mounted in the second ADV well from the front.
    # Y=-31, X=5, Z=8.25 (inches) means to get from the ADVport body to the 
    # ADVport head, go forward along the SMB centerline 5", then go to port 31", and down 8.25" to the head. 
    # (+z is down in ADV body frame of reference). 
    # From dolfyn github web page, body2head_vec is the 3-element vector that specifies the position of the
    # ADV head in the ADV pressure case’s coordinate system
    dat.props['body2head_rotmat'] = np.eye(3)
    dat.props['body2head_vec'] = np.array([5,-31, 8.25]) * 0.0254

    #######
    # Set the declination.
    dat.set_declination(-16.74) 

    #######
    # Calculate and save DOLfYN-defined h,p,r
    odat = dat.orient
    (odat['heading'], odat['pitch'], odat['roll']) = orient2euler(odat['orientmat'])
    # Note dat.orient.raw will have heading, pitch, roll fields that are all 0s
    # because this ADV has an IMU rather than internal compass and tilt sensors.

    # This isn't used by dolfyn, but it may be useful and so we'll 
    # store it here if needed. It is the rotation matrix from the
    # body to the StableMoor buoy reference frame
    # (i.e., it flips y and z)
    dat.props['body2SMB_rotmat'] = np.array([[1, 0, 0],
                        [0, -1, 0],
                        [0, 0, -1]])

    #######
    # Save
    print('Saving port ADV...read_from_VEC_INST_d112.h5')
    dat.to_hdf5(str(outdir / (fname_adv_port + '_read_from_VEC_INST_d112.h5')))
    print('Done saving port ADV ...read_from_VEC_INST_d112.h5')


def write_adv_stbd_csv(outdir=OUTDIR):

    # This block writes some of the variables from the 
    # 5-minute means of motion-corrected starboard
    # ADV data in ETU to .csv files.

    #######
    bdat = dolfyn.load(str(datdir / (fname_adv + '_d112_ETU_5min.h5')))
    bdat['uu'] = bdat['tke_vec'][0]
    bdat['vv'] = bdat['tke_vec'][1]
    bdat['ww'] = bdat['tke_vec'][2]
    bdat['Umag'] = bdat.U_mag
    base.adv_to_csv(bdat,
                    str(outdir / (fname_adv + '_d112_ETU_5min.csv')),
                    info_vars=
                    [
                        ('Horizontal Velocity Mag. [m/s]', 'Umag'),
                        ('water depth [m]', 'water_depth_est'),
                        ('height above bottom [m]', 'ulow_hab'),
                        ('depth below surface [m]', 'depth_v20_16Hz'),
                        ('Turbulence Intensity [-]', 'TI',),
                        ("u'u' [m^2/s^2]", 'uu', ),
                        ("v'v' [m^2/s^2]", 'vv', ),
                        ("w'w' [m^2/s^2]", 'ww', ),
                        ('Temperature [C]', 'env.temp'),
                    ])


if __name__ == '__main__':

    
    read_adv_stbd()
    clean_adv_stbd()
    
    read_RDI()
    clean_RDI()

    correct_motion()
    
    avg_adv_stbd()
    avg_RDI()

    read_adv_port()

    write_adv_stbd_csv()
