'''
Process Western Passage 2017 bottom-mounted Nortek Signature 500 data.
'''

#%%
import numpy as np
import dolfyn
import matplotlib
matplotlib.use('TKAgg') 
from scipy.io.matlab import loadmat
from dolfyn import ADVdata
from dolfyn.data.base import data as dgroup, config
from dolfyn.rotate.signature import euler2orient

from matplotlib import pyplot as plt 
import matplotlib.dates as mdates
from datetime import datetime
import pandas as pd
import zipfile

import base

datdir = base.thisdir
OUTDIR = datdir


#######
# from https://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]


def process_full(outdir=OUTDIR):

    #######
    # Import the .mat files created in Ocean Contour software.
    # dat = loadmat(datdir / 'Burst.VTC.DSEL.AVER_prcntgd50.mat')
    # idat = loadmat(datdir / 'IBurst.DSEL.AVER_prcntgd50.mat')
    dat = loadmat(str(datdir / 'WP2017_Bottom_Sig500_OceanContour_Burst.VTC.DSEL_binmap.AVER.mat'))
    idat = loadmat(str(datdir / 'WP2017_Bottom_Sig500_OceanContour_IBurst.DSEL.AVER.mat'))

    #######
    # Import the .mat file of data from NDBC Station PSBM1 - 8410140 - Eastport, Maine 
    NDBCdat = loadmat(str(datdir / 'WP2017_TideMet_NDBC_PSBM1.mat')) # Data from NDBC Station PSBM1 - 8410140 - Eastport, Maine
    ndbc_time = NDBCdat['dn_met'][:,0] - 366  # shape (29060,)  # Adjust from Matlab time to Python time
    ndbc_airP_hPa = NDBCdat['pair_hPa'][:,0] # shape (29060,) # Air pressure in hPa
    ndbc_airP_Pa = ndbc_airP_hPa * 100 # 1 hPa = 100 Pa # Air pressure in Pa
    # Are there NaN in this time series?
    inan = np.isnan(ndbc_airP_Pa); print(inan.sum()) # 489

    y =ndbc_airP_Pa
    nans, x= nan_helper(y)
    y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    ndbc_airP_Pa2 = y
    inan22 = np.isnan(ndbc_airP_Pa2); print(inan22.sum()) # 0

    #######
    # Assign the filled time series the right variable name for later.
    ndbc_airP_Pa = ndbc_airP_Pa2

    #######
    # dat = loadmat('Burst.VTC.DSEL.AVER_prcntgd50.mat')
    # idat = loadmat('IBurst.DSEL.AVER_prcntgd50.mat')
    # Note, this data is straight from Ocean Contour. 
    # These 5-minute means were created from data collected in ENU.
    # They were bin mapped to correct for tilt, the mask applies to any values
    # collected within 10% of the surface, for amplitudes < 30 dB, and correclations
    # < 50%. The 5-minute means were only generated if more than 50% of the 2 Hz
    # samples were 'good' in the 5 minute (600 sample) period.
    # Need to apply the Data Mask to filter out the bad data. 

    #######
    # This strips away some of the matfile structure to get closer to the data
    d = dat[u'Burst_Data\x00\x00\x00\x00\x00\x00'][0,0]
    i = idat[u'IBurst_Data\x00\x00\x00\x00\x00'][0,0]

    ######
    # These are the pieces of the data structure that are needed for
    # DOLfYN to function properly

    # Initiate the data object
    dout = ADVdata()
    # Setup the properties needed for functionality
    dout.props['coord_sys'] = 'earth'
    dout.props['inst_type'] = 'ADP'
    dout.props['inst_make'] = 'Nortek'
    dout.props['inst_model'] = 'Signature500'
    dout.props['rotate_vars'] = ['vel', ]
    dout['config'] = config(_type='ADP')
    # The TransMatrix is needed if you want to rotate to beam coordinates
    dout['config']['TransMatrix'] = dat[u'Config\x00\x00'][0, 0]['Instrument_burst_beam2xyz'].reshape((4, 4))

    dout['range'] = d['Range'][0] # Range is 1.5, 2.5, 3.5, ... 
                                  # Center of nth cell is blanking distance + n * cell_size = 0.5 + n

    dout['mpltime'] = d['TimeMatlab'][0] - 366
    # MCF: add mpltime for b5, because it's different
    # dout['mpltime_b5'] = i['TimeMatlab'][0] - 366
    # dtime = (dout['mpltime'] - dout['mpltime_b5'])*24*60*60  # All values are 0.125 seconds.
    # The beams1-4 time stamp is 0.125 seconds ahead of the beam5 time stamp. In a 5 minute mean, 
    # that doesn't matter, so we can use mpltime for all 5 beams.

    dout['vel'] = np.stack((d['BinMapVelEast'],
                            d['BinMapVelNorth'],
                            d['BinMapVelUp1'],
                            d['BinMapVelUp2'],
    ))
    dout['vel_b5'] = i['VelBeam5']

    # Copy orientation data over (this isn't actually needed)
    dout['orient'] = dgroup()
    odr = dout['orient']['raw'] = dgroup()
    odr['pitch'] = d['Pitch'][0]
    odr['roll'] = d['Roll'][0]
    odr['heading'] = d['CorrectedHeading'][0]
    # Construct the orientation matrix (this IS NEEDED for rotations)
    dout['orient']['orientmat'] = euler2orient(odr['heading'], odr['pitch'], odr['roll'])

    # End of dolfyn-critical fields
    ######

    ######
    # The rest of this is just copying ancillary data

    # Add percent good field, add altpressure then calc depth from it and NDBC air pressure
    dout['altpress_dbar'] = d['AltimeterPressure']
    dout['altpress_dbar'] = np.transpose(dout['altpress_dbar'])
    dout['altpress_dbar'] = dout['altpress_dbar'][:,0]
    # Correct for the built in pressure offset of 9.5 dbar in the Sig500 data
    pressure_abs_dbar = dout['altpress_dbar'] + 9.5 
    # Convert the Sig500 pressure to Pa
    dout['press_abs_Pa'] = pressure_abs_dbar * 10000  # 10,000 Pa = 1 dbar 
    # Interpolate the air pressure from NDBC to the same time as the Sig500 data.
        # new_y = np.interp(new_x, old_x, old_y)
    dout['press_air_Pa'] = np.interp(dout['mpltime'], ndbc_time, ndbc_airP_Pa)

    # Solve for water depth, h, from 
    # press_absolute = press_air + pressure_water
    pressure_water_Pa = dout['press_abs_Pa'] - dout['press_air_Pa']
    # pressure_water_Pa = rho_kg_m3 *g_m_s2 * h_m 
    rho = 1020
    g = 9.81
    # h = pressure_water/(rho*g)
    dout['water_depth_m'] = pressure_water_Pa/(rho*g) # shape (30220,)

    fignum = 1
    fig = plt.figure(fignum, figsize=[9, 5])
    fig.clf()
    fig, axs = plt.subplots(3, 1, num=fignum, sharex=True)
    axs[0].plot(dout['mpltime'], dout['water_depth_m'], linestyle = 'none', marker = '.')
    axs[0].set_ylabel('depth [m]')
    axs[1].plot(dout['mpltime'], dout['press_abs_Pa'], linestyle = 'none', marker = '.')
    axs[1].set_ylabel('abs pressure\n at ADCP [Pa]')
    axs[2].plot(dout['mpltime'], dout['press_air_Pa'], linestyle = 'none', marker = '.')
    axs[2].set_ylabel('NDBC air \npressure [Pa]')
    # Some funky points at the beginning and end due to funky pressure readings at ADCP.
    # Since I'm going to select chunks of time from May and June later anyway, I won't
    # crop the time series now. 

    dout['pctgd'] = d['PercentGood']

    dout['mask'] = d['DataMask']
    dout['signal'] = dgroup()

    dout['signal']['amp'] = np.stack((
        d['BinMapAmpBeam1'],
        d['BinMapAmpBeam2'],
        d['BinMapAmpBeam3'],
        d['BinMapAmpBeam4'],
    ))
    dout['signal']['amp_b5'] = i['AmpBeam5']

    dout['signal']['corr'] = np.stack((
        d['BinMapCorBeam1'],
        d['BinMapCorBeam2'],
        d['BinMapCorBeam3'],
        d['BinMapCorBeam4'],
    ))
    dout['signal']['corr_b5'] = i['CorBeam5']

    dout['orient']['mag'] = np.stack((
        d['MagnetometerX'],
        d['MagnetometerY'],
        d['MagnetometerZ'],
    ))
    dout['orient']['accel'] = np.stack((
        d['AccelerometerX'],
        d['AccelerometerY'],
        d['AccelerometerZ'],
    ))

    print('Some of the data was copied from the dataset.'
          'The total list of variables available in the .mat data file are:')
    print(d.dtype.names)

    #######
    # Apply the data mask to NaN the 'bad' values.
    ibd = dout['mask'] != 0 # ibd.shape (54,30220)
    # dout['vel'] shape (4,54,30220)
    # dout['vel_b5'] shape (54,30220)

    dout['vel_gd'] = dout['vel'].copy()
    dout['vel_gd'][:,ibd] = np.nan

    dout['vel_b5_gd'] = dout['vel_b5'].copy()
    dout['vel_b5_gd'][ibd] = np.nan

    # Reassign the vel_gd and vel_b5_gd to vel and vel_b5. 
    dout['vel'] = dout['vel_gd']
    dout['vel_b5'] = dout['vel_b5_gd']

    #######
    # Get rid of unneeded variables  
    dout.pop('vel_gd')  
    dout.pop('vel_b5_gd') 

    #######
    # Save
    print('Saving ... masked_d112_ETU.h5')
    dout.to_hdf5(str(outdir / 'WP2017_Bottom_Sig500_OceanContour_masked_d112_ETU.5min.h5'))
    print('Done saving ... masked_d112_ETU.h5')

    #%%
    #######
    #if flag['make May and June subsets']:
    #  From exploring the plots in Ocean Contour, I found two good patches of
    #  data that I will use going forward: 
    #     % Batch 1
    #         % May 1 2017 00:00:00 to May 26 2017 00:00:00
    #             % heading is steady ~ 105, pitch relatively steady 12-15 deg,
    #             % and roll ~ -10.
    #     % Batch 2
    #         % June 3 12:00:00 to June 23 00:00:00
    #             % heading steady ~ 96, pitch ~ 11, and roll ~ -6.7


def process_subsets(outdir=OUTDIR):
    
    #######
    # Load the full time series.
    dat = dolfyn.load(str(datdir / 'WP2017_Bottom_Sig500_OceanContour_masked_d112_ETU.5min.h5'))

    #######
    # Subset for the May chunk
    start_str = '5/1/17 00:00'
    date_object = datetime.strptime(start_str, '%m/%d/%y %H:%M')
    dn_start = mdates.date2num(date_object)
    end_str = '5/26/17 00:00'
    date_object = datetime.strptime(end_str, '%m/%d/%y %H:%M')
    dn_end = mdates.date2num(date_object)

    trange = [dn_start, dn_end ]
    tinds = (trange[0] <= dat.mpltime) & (dat.mpltime <= trange[-1])
    datMay = dat.subset[tinds] 

    #######
    # NaN all the vel data in bins that have less than 95% good data
    # over the course of the time series. This will NaN out the upper bins at high tide.
    # Per IEC 201 guidelines.
    ngd = (~np.isnan(datMay.U_mag)).sum(1) # total events that are NOT nan, per bin. ngd.shape is (54,)
        # There are all NaN in the top 8 bins due to the side-lobe exclusion.
    igd = ngd > 0.95 * datMay.mpltime.size # Boolean. igd.sum() = 26 --> there are 26 bins that have good values 95% of the 5-min events
    datMay.range[igd] # The good data goes from 2.5 to 27.5 m from the face.
    # NaN the vels in the bins that aren't igd
    datMay['vel'][:,~igd,:] = np.nan

    #######
    # Calculate MLLW for the May chunk
    # 288 because 1 day has 288 5 minute means. Reshape puts all the altpress into one column
    # for each day, then finds the minimum for each day, then takes the mean of all the days.
    # The Sig500 data doesn't have a depth field, so I'm using the pressure in dbar.
    # mllw = np.round(datMay['altpress_dbar'].reshape(-1, 288).min(1).mean(), decimals = 1) # 51.8 m when I use the altpressure
    datMay['mllw'] = np.round(datMay['water_depth_m'].reshape(-1, 288).min(1).mean(), decimals = 1) # 51.1 m when I use the water depth.  
    print(datMay['mllw'])

    #######
    # Calculate z_bin (depth, where negative is down from MLLW)
    datMay['z_'] = datMay['range'] - datMay['mllw']

    #######
    # Calculate eta, sea surface height above MLLW
    datMay['eta'] = datMay['water_depth_m'] - datMay['mllw']

    #######
    # Save the May subset.
    print('Saving ... May_d112_ETU.h5')
    datMay.to_hdf5(str(outdir / 'WP2017_Bottom_Sig500_OceanContour_May_d112_ETU.5min.h5'))
    print('Done saving ... May_d112_ETU.h5')

    #######
    # Subset the June chunk
    start_str = '6/4/17 00:00' 
    date_object = datetime.strptime(start_str, '%m/%d/%y %H:%M')
    dn_start = mdates.date2num(date_object)
    end_str = '6/23/17 00:00'
    date_object = datetime.strptime(end_str, '%m/%d/%y %H:%M')
    dn_end = mdates.date2num(date_object)

    trange = [dn_start, dn_end ]
    tinds = (trange[0] <= dat.mpltime) & (dat.mpltime <= trange[-1])
    datJune = dat.subset[tinds]  

    #######
    # NaN all the vel data in bins that have less than 95% good data
    # over the course of the time series. This will NaN out the upper bins at high tide.
    # Per IEC 201 guidelines.
    ngd = (~np.isnan(datJune.U_mag)).sum(1) # total events that are NOT nan, per bin. ngd.shape is (54,)
    igd = ngd > 0.95 * datJune.mpltime.size # Boolean. igd.sum() = 30 --> there are 30 bins that have good values 95% of the 5-min events
    datJune.range[igd] # The good data goes from 2.5 to 31.5 m from the face.
    # NaN the vels in the bins that aren't igd
    datJune['vel'][:,~igd,:] = np.nan

    #######
    # Calculate MLLW for the June chunk
    # 288 because 1 day has 288 5 minute means. Reshape puts all the altpress into one column
    # for each day, then finds the minimum for each day, then takes the mean of all the days.
    # The Sig500 data doesn't have a depth field, so I'm using the pressure in dbar.
    # mllw = np.round(datJune['altpress_dbar'].reshape(-1, 288).min(1).mean(), decimals = 1) # 51.8 m when I use the altpressure
    datJune['mllw'] = np.round(datJune['water_depth_m'].reshape(-1, 288).min(1).mean(), decimals = 1) # 51.9 m when I use the water depth.  
    print(datJune['mllw'])

    #######
    # Calculate z_bin (depth, where negative is down from MLLW)
    datJune['z_'] = datJune['range'] - datJune['mllw']

    #######
    # Calculate eta, sea surface height above MLLW
    datJune['eta'] = datJune['water_depth_m'] - datJune['mllw']

    #######
    # Save the June subset.
    print('Saving ... June_d112_ETU.h5')
    datJune.to_hdf5(str(outdir / 'WP2017_Bottom_Sig500_OceanContour_June_d112_ETU.5min.h5'))
    print('Done saving ... June_d112_ETU.h5')


def write_csv(outdir=OUTDIR):

    # This block writes some of the variables to .csv files
    # following def to_csv above.

    #######
    # Load data
    datMay = dolfyn.load(str(datdir / ('WP2017_Bottom_Sig500_OceanContour_May_d112_ETU.5min.h5')))
    datJune = dolfyn.load(str(datdir / ('WP2017_Bottom_Sig500_OceanContour_June_d112_ETU.5min.h5')))

    fnameMay = 'WP2017_Bottom_Sig500_OceanContour_May_d112_ETU.5min'
    fnameJune = 'WP2017_Bottom_Sig500_OceanContour_June_d112_ETU.5min'

    datMay.z = datMay.z_
    datJune.z = datJune.z_
    
    # Write to .csv
    base.to_csv(datMay, str(outdir / fnameMay),
                info_vars=[('depth_m', 'water_depth_m'),
                           # No temperature data here
                           #('temp_C', 'env.temperature_C'),
                ],
    )
    base.to_csv(datJune, str(outdir / fnameJune),
                info_vars=[('depth_m', 'water_depth_m'),
                           # No temperature data here
                           #('temp_C', 'env.temperature_C'),
                ],
    )
    
    # %%

if __name__ == '__main__':

    process_full()
    process_subsets()
    write_csv()

