'''
Process Western Passage 2017 bottom-mounted RDI Workhorse 600 data.
'''
#%% 
import numpy as np
import dolfyn
from dolfyn.rotate.base import orient2euler
import base

datdir = base.thisdir
fname = 'WP2017_Bottom_RDI600'
OUTDIR = base.thisdir

def read_raw(outdir=OUTDIR):

    # Read the raw RDI file from the bottom-mounted Workhorse and saves it as .h5      
    dat = dolfyn.read(str(datdir / (fname + '.000')), userdata = False) 
    print('Done reading the .000 file')
    # This file does NOT have a userdata.json file

    # Set the declination.
    # ORPC Deployed Locations figure shows bottom RDI at 44deg55.209 N, 66deg59.349 W
    # https://www.rapidtables.com/convert/number/degrees-minutes-seconds-to-degrees.html
    # 44.92015 N, 66.98915 W
    # https://www.ngdc.noaa.gov/geomag/calculators/magcalc.shtml#declination
   
    dat.set_declination(-16.74) 

    # Calc and save DOLfYN-defined h,p,r
    odat = dat.orient
    (odat['heading'], odat['pitch'], odat['roll']) = orient2euler(odat['orientmat'])

    # Adjust timestamp from UTC-4 to UTC!
    dat['mpltime'] += 4/24

    # Save
    print('Saving  ...raw_ETU_d112.h5')
    dat.to_hdf5(str(outdir / (fname +'_raw_ETU_d112.h5')))
    print('Done saving ...raw_ETU_d112.h5')


def process(outdir=OUTDIR):

    # This block cleans the data.

    #######
    # Load data
    dat = dolfyn.load(str(datdir / (fname + '_raw_ETU_d112.h5')))

    #######
    # Crop the data.
    dat = dat.subset[slice(7, 25404)]

    #######
    # Apply corr filter.
    # WorkHorse Commands and Output Data Format.pdf pg. 84 indicates the Low Correlation Threshold
    # default is 64. 

    # See how much data gets cut when this threshold is employed: dat.signal.corr[ivel] < 128 (and < 64).
    # Assign temporary variable.
    dat['vel_gd'] = dat.vel.copy()

    for ivel in range(3):
        ibad = (dat.signal.corr[ivel] < 64) 
        dat.vel_gd[ivel, ibad] = np.nan

    iNotnan2 = ~np.isnan(dat.vel_gd); print(iNotnan2.sum()) # 5269438
    inan2 = np.isnan(dat.vel_gd); print(inan2.sum()) # 114726
    tots2 = iNotnan2.sum() + inan2.sum()
    prcnt_nan2 = inan2.sum() / tots2 * 100 # When the threshold is 128, 74% of the data
                                            # gets NaN'd. And the next plot has too little data.
                                            # When the threshold is 64, 2.1 % of the data gets
                                            # NaN'd.

    #######
    # if flag['raw prcnt gd']:
    # Plot percent goods
    # The percent good fields are populated differently than you
    # might expect (i.e. that they mean the % of good values) because the 
    # coordinate system was set to Earth.
    # Field 1 is the % of good 3-beam solutions. Since the best solutions
    # are 4 beam solutions (Field 4), high values here
    # are good if Field 4 values are low.
    # Field 2 is the % of transformations rejected because the error
    # velocity was over the threshold (2000 mm/s)
    # Field 3 is the % where more than 1 beam are bad in a bin.
    # Field 4 is the % of 4-beam solutions. Best data.

    # NaN more using percent good
    ibad = (dat.signal.prcnt_gd[0] + dat.signal.prcnt_gd[3] < 90)
    for ivel in range(3):
        dat.vel_gd[ivel, ibad] = np.nan

    iNotnan3 = ~np.isnan(dat.vel_gd); print(iNotnan3.sum()) 
    inan3 = np.isnan(dat.vel_gd); print(inan3.sum()) 
    tots3 = iNotnan3.sum() + inan3.sum()
    prcnt_nan3 = inan3.sum() / tots3 * 100 # 9.7 percent

    # Assign dat.vel_gd to dat.vel and pop dat.vel_gd
    dat.vel = dat.vel_gd
    dat.pop('vel_gd')

    #######
    # Calculate MLLW. 
    # The idea here is to determine the mean minimum water depth (how much water is above the RDI), then
    # create a new variable 'z' that is the bin ranges minus the mean low water depth (in most cases will be a negative number); will be the 
    # distance from surface to bin (negative down).
    # And create new variable 'eta' to subtract MLLW from the height of water above the RDI; leaves sea surface height.
    # Why only to dat.depth_m[:25200]? Because that's 240 samples/day * 105 days. dat is 105.82 days long
    # What is happening here? find mean minimum depth, but what's the reshape part doing? What's the 240 for? Answer: It's taking
    # the one data point every 6 minutes, and making 
    # shape (105, 240) There are 105 rows, each with 240 col; there are 105 days, each with 240 depths (one every 6 min.)
    dat['mllw'] = np.round(dat.depth_m[:25200].reshape(-1, 240).min(1).mean(), decimals = 1)  # 47.2 # Take the min depth for each of the 105 days, then take the mean of that.
                                                        # That means that on average, there is at least 47.15 m of water over the RDI.
    print(dat['mllw'])


    dat['z_'] = dat.range - dat['mllw']  # z is depth (negative down), relative to MLLW. So the range was the distance from the RDI face. Now, z is the depth of water
    # from the surface (at MLLW), to the the location where the measurement is made. NOT a time series. Same size as dat.range.
    # 'eta' is sea surface height above MLLW.
    dat['eta'] = dat.depth_m - dat['mllw']

    #######
    # Mask the values above 0.90 of the water depth.
    # The Broadband Primer_Dec06.pdf, pg. 38. Rmax = D cos(theta) where
    # Rmax is maximum range for acceptable data, D is distance from ADCP to surface (or bottom),
    # theta is angle of beam reltive to vertical (in this case 20 degrees). 
    # cos(theta) = .94
    goods = dat.range[:, None] < dat.depth_m[None, :] * 0.9
    dat.vel[:, ~goods] = np.NaN

    iNotnan4 = ~np.isnan(dat.vel); print(iNotnan4.sum()) # 
    inan4 = np.isnan(dat.vel); print(inan4.sum()) # 
    tots4 = iNotnan4.sum() + inan4.sum()
    prcnt_nan4 = inan4.sum() / tots4 * 100 # 19.8

    #######
    # NaN all the vel data in bins that have less than 95% good data
    # over the course of the time series. This will NaN out the upper bins at high tide.
    # Per IEC 201 guidelines.
    ngd = (~np.isnan(dat.U_mag)).sum(1) # Total events that are NOT nan, per bin. ngd.shape (53,)
    igd = ngd > 0.95 * dat.mpltime.size # Boolean. The last 17 bins are False. First good bin is -10 m below MLLW
    dat.vel[:, ~igd] = np.NaN
    dat.range[~igd] = np.NaN # Good range is from 2.11 to 37.11 m from face.
    dat.z_[~igd] = np.NaN

    iNotnan5 = ~np.isnan(dat.vel); print(iNotnan5.sum()) # 3651458
    inan5 = np.isnan(dat.vel); print(inan5.sum()) # 1732706
    tots5 = iNotnan5.sum() + inan5.sum()
    prcnt_nan5 = inan5.sum() / tots5 * 100 # 32.2

    #######
    # Save
    print('Saving ...cln_ETU_d112.h5')
    dat.to_hdf5(str(outdir / (fname + '_cln_ETU_d112.h5')))
    print('Done saving ...cln_ETU_d112.h5')


def write_csv(outdir=OUTDIR):

    # This block writes some of the variables to .csv files

    #######
    # Load data
    dat = dolfyn.load(str(datdir / (fname + '_cln_ETU_d112.h5')))

    # Write to .csv
    seabed_to_RDI_head_m = 0.5
    dat.depth_m += seabed_to_RDI_head_m
    dat.z = dat.z_
    base.to_csv(dat, str(outdir / fname),
                info_vars=[('bottom_m', 'depth_m'),
                           ('temp_C', 'env.temperature_C'),
                ])


if __name__ == '__main__':
    # If this file is run as a script, execute the following. 
    read_raw()
    process()
    write_csv()

