import dolfyn
import matplotlib.pyplot as plt
import numpy as np
import glob
import json
import dolfyn.tools.misc as tbx
import gis as gistools
import pandas as pd
from datetime import datetime
import zipfile
import base
import pathlib

datdir = base.thisdir

allfiles = list(datdir.glob('*.PD0'))

with (datdir / 'WP2017_MobileTransect_angles.json').open('rb') as ang_file:
    angles = json.load(ang_file)


proj_angle = 45.0
proj = gistools.simple_proj([-66.990367, 44.919750],
                            proj_angle)

depth_transducer = 0.6  # m


def rotate_vel(dat, ang):
    U = dat[0] + 1j * dat[1]
    U *= np.exp(1j * ang)
    return U.real, U.imag


def calc_xy(dat, ):
    """Calculate position, for calculating speed."""

    
    dt = np.median(np.diff(dat.mpltime))
    Uc = dat['bt_vel'][0] + 1j * dat['bt_vel'][1]
    bads = np.abs(Uc) > 10
    Uc[bads] = np.NaN
    dat['bt_vel'][2][bads] = np.NaN
    tbx.fillgaps(dat['bt_vel'][2], extrapFlg=True)
    tbx.fillgaps(Uc, extrapFlg=True)
    dat['bt_vel'][:2] = Uc.real, Uc.imag
    Xbt = np.cumsum(Uc) * 24 * 3600 * dt

    llind = np.nonzero(~np.isnan(dat.orient.glongitude) &
                       (dat.orient.glongitude != 0))[0][0]
    Xbt *= np.exp(1j * -(90 - proj_angle) * np.pi / 180)
    xy0 = proj((dat.orient.glongitude[llind],
                dat.orient.glatitude[llind]))
    xy0 = xy0[0] + 1j * xy0[1]
    Xbt += xy0 - Xbt[llind]
    return Xbt


def process_all(outdir=None, figdir=None, transect_coords=False,
                n_max=None):
    """Process the transect files.
    
    Parameters
    ==========
    outdir : {None (default) or str}
       If None, than outdir is the same as datdir (set above).

    figdir : {None (default), or str, or pathlib.Path}
       If None, no figures are generated.

    transect_coords {bool}
       Rotate the data to be in the coordinate system of the transect
       (y,v are along transect, x,u are normal to it). This can be
       used to calculate fluxes, but BEWARE: this rotations is not
       tracked in the DOLFyN coordinate system!
    """
    if figdir is not None and not isinstance(figdir, pathlib.Path):
        figdir = pathlib.Path(figdir)
    
    if outdir is None:
        outdir = datdir
    elif not isinstance(outdir, pathlib.Path):
        outdir = pathlib.Path(outdir)
    
    for iii in range(len(allfiles)):

        file = allfiles[iii].name.split('.')[0]
        print('\n**** PROCESSING FILE: {} ****'.format(file))

        dat = dolfyn.read(str(datdir / (file + '.PD0')))
        dat.rotate2('earth', inplace=True)
        dat.mpltime = dolfyn.data.time.time_array(dat.mpltime)

        if transect_coords:
            # Rotate to the transect's axes
            dat['bt_vel'][:2] = rotate_vel(dat['bt_vel'][:2],
                                           angles[file] -
                                           proj_angle * np.pi / 180)
            dat['vel'][:2] = rotate_vel(dat['vel'][:2],
                                        angles[file] -
                                        proj_angle * np.pi / 180)
            X = calc_xy(dat)
            dat['x'] = X.real
            dat['y'] = X.imag
            
        # Subtract bottom (i.e., ship motion) velocity.
        dat['vel'] -= dat['bt_vel'][:, None, :]

        # Remove bad values in bt_range
        dat.bt_range[dat.bt_range < 20] = np.NaN
        # And fill the gaps
        tbx.fillgaps(dat.bt_range, maxgap=21,
                     dim=1, extrapFlg=True)

        # NaN values with low correlation
        dat['vel'][:, (dat['signal.corr'] < 90).any(0)] = np.NaN

        # Set the bottom depth
        dat['bottom'] = np.nanmin(dat.bt_range, 0) + depth_transducer

        # Set the depth
        dat['z'] = dat.range + depth_transducer

        dat.vel[:, dat.range[:, None] >
                (dat.bottom[None, :] * 0.85)] = np.NaN

        # Apply an 11x11 median filter
        bads = np.zeros(dat['vel'][0].shape, dtype='bool')
        for iax in range(3):
            bads |= np.isnan(tbx.medfiltnan(dat['vel'][iax],
                                            [11, 11], 0.95))
        dbot = np.abs(dat.range[:, None] - dat.bottom[None, :])
        bot_inds = dbot == dbot.min(0)
        for iax in range(3):
            # Set bad values
            dat['vel'][iax][bads] = np.NaN
            # Fill gaps temporally
            tbx.fillgaps(dat['vel'][iax], maxgap=5, dim=0)
            # ... and vertically
            tbx.fillgaps(dat['vel'][iax, :20], maxgap=5000, dim=1)
            tbx.fillgaps(dat['vel'][iax, 20:], maxgap=21, dim=1)
            # Force bottom velocity to zero
            dat['vel'][iax][bot_inds] = 0
            tbx.fillgaps(dat['vel'][iax], dim=0)
            # Apply another (now 9x9) median filter
            dat['vel'][iax] = tbx.medfiltnan(dat['vel'][iax],
                                             [9, 9])
            # Double check that the data is NaN below the bottom
            dat.vel[:, dat.range[:, None] >
                    (dat.bottom[None, :])] = np.NaN
            # Fill it one last time...
            dat['vel'][iax][bot_inds] = 0
            tbx.fillgaps(dat['vel'][iax], dim=0)

        dat.to_hdf5(str(outdir / (file + '.h5')))

        base.to_csv(dat, str(outdir / file),
                    info_vars=[
                        ('Lat', 'orient.glatitude'),
                        ('Long', 'orient.glongitude'),
                        ('bottom_m', 'bottom'),
                        ('temp_C', 'env.temperature_C'),
                    ],
        )

        if figdir is not None and figdir.is_dir():
            # Set figdir (defined above) to something to output figures.

            cmap = plt.get_cmap('RdBu_r')
            cmap.set_bad('0.7')

            fig1 = plt.figure(10, figsize=[6, 8])
            fig1.clf()

            fig1, axs = plt.subplots(4, 1,
                                     sharex=True,
                                     sharey=True, num=10,
                                     gridspec_kw=dict(
                                         left=0.1, right=0.8
                                     ))

            for iax in range(3):
                ax = axs[iax]
                pcol = ax.pcolormesh(dat.x, dat.range,
                                     np.ma.masked_invalid(dat['vel'][iax]),
                                     vmin=-3, vmax=3,
                                     cmap=cmap)

                ax.plot(dat.x, dat.bottom, 'k-')
                bbax = ax.get_position().bounds
                cbax = fig1.add_axes([bbax[0] + bbax[2] + 0.04,
                                      bbax[1], 0.04, bbax[3]])
                cbar = plt.colorbar(pcol, cax=cbax)
                cbar.ax.set_ylabel('[m/s]')
                ax.set_title('{} Velocity'.format(vel_comp[iax]), size='medium')

            ax = axs[-1]
            pcol = ax.pcolormesh(dat.x, dat.range,
                                 np.ma.masked_invalid(
                                     (dat['vel'][:3] ** 2).sum(0) ** 0.5
                                 ),
                                 vmin=-3, vmax=3,
                                 cmap=cmap)
            ax.plot(dat.x, dat.bottom, 'k-')
            ax.set_title('Velocity Magnitude')

            ax.set_ylim([115, 0])
            ax.set_ylabel('Depth [m]')
            ax.set_xlabel('x [m]')
            ax.set_xlim([-100, 600])

            fig1.savefig(figdir + file + '_transect_picture01.png')

            fig2 = plt.figure(12)
            fig2.clf()
            ax2 = plt.gca()

            ax2.plot(dat.x, dat.y)
            ax2.set_aspect('equal')

        if n_max is not None and iii >= n_max:
            break
            
if __name__ == '__main__':
    # If this file is run as a script.

    # Do the data processing
    process_all()
