import dolfyn
import pathlib
import filecmp
import zipfile
import base
import warnings
import pyDictH5.base

# The np.isclose test is too strict (1e-6) for some of the ADV data, so relax it here.
pyDictH5.base.arrayEQ_tols['atol'] = 1e-5
# Print differences between data files.
pyDictH5.base.debug_level = 10

test_outdir = base.thisdir / 'test'

def compare_h5(testdir, srcdir, glob_str='*.h5'):
    testdir = pathlib.Path(testdir)
    srcdir = pathlib.Path(srcdir)

    try:
        _compare_file_lists(testdir.glob(glob_str),
                            srcdir.glob(glob_str))
    except AssertionError:
        warnings.warn("The files lists do not match")
    
    for tf in testdir.glob(glob_str):
        print("Comparing {} ... ".format(tf.name), end='')
        td = dolfyn.load(str(tf))
        sd = dolfyn.load(str(srcdir / tf.name))
        try:
            assert td == sd
        except AssertionError:
            warnings.warn("File {} does not match!".format(tf.name))
        else:
            print(" OK!")


def _compare_file_lists(list1, list2):
    """list1 and list2 must be lists or iterables of <pathlib.Path>
    objects.
    """
    s1 = {tf.name for tf in list1}
    s2 = {tf.name for tf in list2}
    assert (s1 == s2), "The list of files do not match!"


def compare_csvzip(testdir, srcdir, glob_str='*.csv.zip'):
    testdir = pathlib.Path(testdir)
    srcdir = pathlib.Path(srcdir)
    test_files = list(testdir.glob(glob_str))
    try:
        _compare_file_lists(test_files, list(srcdir.glob(glob_str)))
    except AssertionError:
        warnings.warn("The files lists do not match")
    for tf in test_files:
        print("Comparing {} ...".format(tf.name), end='')
        if _compare_zip(tf, srcdir / tf.name):
            print(" OK!")
        else:
            print("The file {} did not match.".format(tf))
        

def _compare_zip(fname1, fname2):
    # Compare the CHECKSUMs of the files in the zipfiles.
    with zipfile.ZipFile(str(fname1), 'r') as zf:
        zinfo1 = {zi.filename: zi.CRC for zi in zf.infolist()}
    with zipfile.ZipFile(str(fname2), 'r') as zf:
        zinfo2 = {zi.filename: zi.CRC for zi in zf.infolist()}
    if zinfo1 != zinfo2:
        for ky in zinfo1:
            if zinfo1[ky] != zinfo2[ky]:
                warnings.warn("The files {} does not match.".format(ky))
        return False
    else:
        return True


if __name__ == '__main__':
    # Script.
    import process_transects
    import process_btmSig500
    import process_btmWkhs
    import process_STTM
    
    test_outdir.mkdir(exist_ok=True)

    ##### Transect Data #####
    # Generate the processed data files, and write them to test_outdir
    process_transects.process_all(outdir=test_outdir)
    # Now confirm that they are identical to the ones in datdir
    compare_h5(testdir=test_outdir,
               srcdir=process_transects.datdir,
               glob_str='WP2017_MobileTransect_RDI300*.h5')
    compare_csvzip(testdir=test_outdir,
                   srcdir=process_transects.datdir,
                   glob_str='WP2017_MobileTransect_RDI300*.csv.zip')

    ##### Bottom Signature 500 #####
    # Generate processed Signature 500 files
    process_btmSig500.process_full(outdir=test_outdir)
    process_btmSig500.process_subsets(outdir=test_outdir)
    # Confirm that files match.
    compare_h5(testdir=test_outdir,
               srcdir=process_btmSig500.datdir,
               glob_str='WP2017_Bottom_Sig500*.h5')

    ##### Bottom Workhorse 600 #####
    # Generate processed Workhorse 600 files
    process_btmWkhs.read_raw(outdir=test_outdir)
    process_btmWkhs.process(outdir=test_outdir)
    # Confirm that files match.
    compare_h5(testdir=test_outdir,
               srcdir=process_btmWkhs.datdir,
               glob_str='WP2017_Bottom_RDI600*.h5')

    
    ##### STTM #####
    process_STTM.read_adv_stbd(outdir=test_outdir)
    process_STTM.clean_adv_stbd(outdir=test_outdir)
    
    process_STTM.read_RDI(outdir=test_outdir)
    process_STTM.clean_RDI(outdir=test_outdir)

    process_STTM.correct_motion(outdir=test_outdir)
    
    process_STTM.avg_adv_stbd(outdir=test_outdir)
    process_STTM.avg_RDI(outdir=test_outdir)

    process_STTM.read_adv_port(outdir=test_outdir)
    
    # Confirm that files match.
    compare_h5(testdir=test_outdir,
               srcdir=process_STTM.datdir,
               glob_str='WP2017_STTM*.h5')
